@@ -162,7 +162,21 @@ class Storage(storage.Storage):
         return changes or len(to_delete + to_uncomplete) > 0
 
     def __get_categories(self):
-        return self.fetchall("SELECT * FROM categories ORDER BY category_order")
+        return self.fetchall("SELECT id, name FROM categories ORDER BY lower(name)")
+
+    def __update_activity(self, id, name, category_id):
+        query = """
+                   UPDATE activities
+                       SET name = ?,
+                           search_name = ?,
+                           category_id = ?
+                     WHERE id = ?
+        """
+        self.execute(query, (name, name.lower(), category_id, id))
+
+        affected_ids = [res[0] for res in self.fetchall("select id from facts where activity_id = ?", (id,))]
+        self.__remove_index(affected_ids)
+
 
     def __change_category(self, id, category_id):
         # first check if we don't have an activity with same name before us
@@ -186,54 +200,45 @@ class Storage(storage.Storage):
             self.__remove_activity(id)
 
         else: #just moving
-            query = "SELECT max(activity_order) + 1 FROM activities WHERE category_id = ?"
-            max_order = self.fetchone(query, (category_id, ))[0] or 1
-
             statement = """
                        UPDATE activities
-                          SET category_id = ?, activity_order = ?
+                          SET category_id = ?
                         WHERE id = ?
             """
 
-            self.execute(statement, (category_id, max_order, id))
+            self.execute(statement, (category_id, id))
+
+        affected_ids = [res[0] for res in self.fetchall("select id from facts where activity_id = ?", (id,))]
+        if existing_activity:
+            affected_ids.extend([res[0] for res in self.fetchall("select id from facts where activity_id = ?", (existing_activity['id'],))])
+        self.__remove_index(affected_ids)
 
         return True
 
     def __add_category(self, name):
-        order = self.fetchone("select max(category_order) + 1  from categories")[0] or 1
         query = """
-                   INSERT INTO categories (name, category_order)
+                   INSERT INTO categories (name, search_name)
                         VALUES (?, ?)
         """
-        self.execute(query, (name, order))
+        self.execute(query, (name, name.lower()))
         return self.__last_insert_rowid()
 
     def __update_category(self, id,  name):
         if id > -1: # Update, and ignore unsorted, if that was somehow triggered
             update = """
                        UPDATE categories
-                           SET name = ?
+                           SET name = ?, search_name = ?
                          WHERE id = ?
             """
-            self.execute(update, (name, id))
-
-    def __move_activity(self, source_id, target_order, insert_after = True):
-        statement = "UPDATE activities SET activity_order = activity_order + 1"
-
-        if insert_after:
-            statement += " WHERE activity_order > ?"
-        else:
-            statement += " WHERE activity_order >= ?"
-
-        self.execute(statement, (target_order, ))
-
-        statement = "UPDATE activities SET activity_order = ? WHERE id = ?"
-
-        if insert_after:
-            self.execute(statement, (target_order + 1, source_id))
-        else:
-            self.execute(statement, (target_order, source_id))
+            self.execute(update, (name, name.lower(), id))
 
+        affected_query = """
+            SELECT id
+              FROM facts
+             WHERE activity_id in (SELECT id FROM activities where category_id=?)
+        """
+        affected_ids = [res[0] for res in self.fetchall(affected_query, (id,))]
+        self.__remove_index(affected_ids)
 
 
     def __get_activity_by_name(self, name, category_id = None, resurrect = True):
@@ -263,7 +268,8 @@ class Storage(storage.Storage):
             res = self.fetchone(query, (_("Unsorted"), name, ))
 
         if res:
-            res = dict(res)
+            keys = ('id', 'name', 'deleted', 'category')
+            res = dict([(key, res[key]) for key in keys])
             res['deleted'] = res['deleted'] or False
 
             # if the activity was marked as deleted, resurrect on first call
@@ -420,6 +426,8 @@ class Storage(storage.Storage):
                                           end_time, start_time, end_time, start_time))
 
         for fact in conflicts:
+            # won't eliminate as it is better to have overlapping entries than loosing data
+
             # split - truncate until beginning of new entry and create new activity for end
             if fact["start_time"] < start_time < fact["end_time"] and \
                fact["start_time"] < end_time < fact["end_time"]:
@@ -441,13 +449,6 @@ class Storage(storage.Storage):
                                       WHERE fact_id = ?"""
                 self.execute(tag_update, (new_fact_id, fact["id"])) #clone tags
 
-            #eliminate
-            elif fact["end_time"] and \
-                 start_time < fact["start_time"] < end_time and \
-                 start_time < fact["end_time"] < end_time:
-                logging.info("eliminating %s" % fact["name"])
-                self.__remove_fact(fact["id"])
-
             # overlap start
             elif start_time < fact["start_time"] < end_time:
                 logging.info("Overlapping start of %s" % fact["name"])
@@ -477,7 +478,7 @@ class Storage(storage.Storage):
         tags = [tag.strip() for tag in tags.split(",") if tag.strip()]  # split by comma
         tags = tags or activity.tags # explicitly stated tags take priority
 
-        tags = self.GetTagIds(tags) #this will create any missing tags too
+        tags = [dict(zip(('id', 'name', 'autocomplete'), row)) for row in self.GetTagIds(tags)] #this will create any missing tags too
 
 
         if category_name:
@@ -595,6 +596,7 @@ class Storage(storage.Storage):
         params = [(fact_id, tag["id"]) for tag in tags]
         self.execute(insert, params)
 
+        self.__remove_index([fact_id])
         return fact_id
 
     def __last_insert_rowid(self):
@@ -611,6 +613,16 @@ class Storage(storage.Storage):
 
 
     def __get_facts(self, date, end_date = None, search_terms = ""):
+        from configuration import conf
+        day_start = conf.get("day_start_minutes")
+        day_start = dt.time(day_start / 60, day_start % 60)
+
+        split_time = day_start
+        datetime_from = dt.datetime.combine(date, split_time)
+
+        end_date = end_date or date
+        datetime_to = dt.datetime.combine(end_date, split_time) + dt.timedelta(days = 1)
+
         query = """
                    SELECT a.id AS id,
                           a.start_time AS start_time,
@@ -627,54 +639,18 @@ class Storage(storage.Storage):
                     WHERE (a.end_time >= ? OR a.end_time IS NULL) AND a.start_time <= ?
         """
 
-        # let's see what we can do with search terms
-        # we will be looking in activity names, descriptions, categories and tags
-        # comma will be treated as OR
-        # space will be treated as AND or possible join
+        if search_terms:
+            # check if we need changes to the index
+            self.__check_index(datetime_from, datetime_to)
 
-
-        # split by comma and then by space and remove all extra spaces
-        or_bits = [[term.strip().lower().replace("'", "''") #striping removing case sensitivity and escaping quotes in term
-                          for term in terms.strip().split(" ") if term.strip()]
-                          for terms in search_terms.split(",") if terms.strip()]
-
-        def all_fields(term):
-            return """(lower(a.description) like '%%%(term)s%%'
-                       or lower(b.name) = '%(term)s'
-                       or lower(c.name) = '%(term)s'
-                       or lower(e.name) = '%(term)s' )""" % dict(term = term)
-
-        if or_bits:
-            search_query = "1<>1 " # will be building OR chain, so start with a false
-
-            for and_bits in or_bits:
-                if len(and_bits) == 1:
-                    and_query = all_fields(and_bits[0])
-                else:
-                    and_query = "1=1 "  # will be building AND chain, so start with a true
-                    # if we have more than one word, go for "(a and b) or ab"
-                    # to match two word tags
-                    for bit1, bit2 in zip(and_bits, and_bits[1:]):
-                        and_query += "and (%s and %s) or %s" % (all_fields(bit1),
-                                                                all_fields(bit2),
-                                                                all_fields("%s %s" % (bit1, bit2)))
-
-                search_query = "%s or (%s) " % (search_query, and_query)
-
-            query = "%s and (%s)" % (query, search_query)
+            search_terms = search_terms.replace('\\', '\\\\').replace('%', '\\%').replace('_', '\\_').replace("'", "''")
+            query += """ AND a.id in (SELECT id
+                                        FROM fact_index
+                                       WHERE fact_index MATCH '%s')""" % search_terms
 
 
 
         query += " ORDER BY a.start_time, e.name"
-        end_date = end_date or date
-
-        from configuration import conf
-        day_start = conf.get("day_start_minutes")
-        day_start = dt.time(day_start / 60, day_start % 60)
-
-        split_time = day_start
-        datetime_from = dt.datetime.combine(date, split_time)
-        datetime_to = dt.datetime.combine(end_date, split_time) + dt.timedelta(days = 1)
 
         facts = self.fetchall(query, (_("Unsorted"),
                                       datetime_from,
@@ -732,23 +708,20 @@ class Storage(storage.Storage):
                       "DELETE FROM facts where id = ?"]
         self.execute(statements, [(fact_id,)] * 2)
 
+        self.__remove_index([fact_id])
+
     def __get_category_activities(self, category_id):
         """returns list of activities, if category is specified, order by name
            otherwise - by activity_order"""
         query = """
-                   SELECT a.id, a.name, a.activity_order, a.category_id, b.name as category
+                   SELECT a.id, a.name, a.category_id, b.name as category
                      FROM activities a
                 LEFT JOIN categories b on coalesce(b.id, -1) = a.category_id
                     WHERE category_id = ?
                       AND deleted is null
+                 ORDER BY lower(a.name)
         """
 
-        # unsorted entries we sort by name - others by ID
-        if category_id == -1:
-            query += "ORDER BY lower(a.name)"
-        else:
-            query += "ORDER BY a.activity_order"
-
         return self.fetchall(query, (category_id, ))
 
 
@@ -762,11 +735,12 @@ class Storage(storage.Storage):
                 LEFT JOIN categories b ON coalesce(b.id, -1) = a.category_id
                 LEFT JOIN facts f ON a.id = f.activity_id
                     WHERE deleted IS NULL
-                      AND lower(a.name) LIKE ? ESCAPE '\\'
+                      AND a.search_name LIKE ? ESCAPE '\\'
                  GROUP BY a.id
                  ORDER BY max(f.start_time) DESC, lower(a.name)
                     LIMIT 50
         """
+        search = search.lower()
         search = search.replace('\\', '\\\\').replace('%', '\\%').replace('_', '\\_')
         activities = self.fetchall(query, (u'%s%%' % search, ))
 
@@ -788,18 +762,20 @@ class Storage(storage.Storage):
     def __remove_category(self, id):
         """move all activities to unsorted and remove category"""
 
+        affected_query = """
+            SELECT id
+              FROM facts
+             WHERE activity_id in (SELECT id FROM activities where category_id=?)
+        """
+        affected_ids = [res[0] for res in self.fetchall(affected_query, (id,))]
+
         update = "update activities set category_id = -1 where category_id = ?"
         self.execute(update, (id, ))
 
         self.execute("delete from categories where id = ?", (id, ))
 
+        self.__remove_index(affected_ids)
 
-    def __swap_activities(self, id1, priority1, id2, priority2):
-        """ swaps nearby activities """
-        # TODO - 2 selects and 2 updates is wrong we could live without selects
-        self.execute(["update activities set activity_order = ? where id = ?",
-                      "update activities set activity_order = ? where id = ?"],
-                      [(priority1, id2), (priority2, id1)])
 
     def __add_activity(self, name, category_id = None, temporary = False):
         # first check that we don't have anything like that yet
@@ -809,7 +785,6 @@ class Storage(storage.Storage):
 
         #now do the create bit
         category_id = category_id or -1
-        new_order = self.fetchone("select max(activity_order) + 1  from activities")[0] or 1
 
         deleted = None
         if temporary:
@@ -817,20 +792,57 @@ class Storage(storage.Storage):
 
 
         query = """
-                   INSERT INTO activities (name, category_id, activity_order, deleted)
+                   INSERT INTO activities (name, search_name, category_id, deleted)
                         VALUES (?, ?, ?, ?)
         """
-        self.execute(query, (name, category_id, new_order, deleted))
+        self.execute(query, (name, name.lower(), category_id, deleted))
         return self.__last_insert_rowid()
 
-    def __update_activity(self, id, name, category_id):
-        query = """
-                   UPDATE activities
-                       SET name = ?,
-                           category_id = ?
-                     WHERE id = ?
-        """
-        self.execute(query, (name, category_id, id))
+    def __remove_index(self, ids):
+        """remove affected ids from the index"""
+        if not ids:
+            return
+
+        ids = ",".join((str(id) for id in ids))
+        self.execute("DELETE FROM fact_index where id in (%s)" % ids)
+
+
+    def __check_index(self, start_date, end_date):
+        """check if maybe index needs rebuilding in the time span"""
+        index_query = """SELECT id
+                           FROM facts
+                          WHERE (end_time >= ? OR end_time IS NULL)
+                            AND start_time <= ?
+                            AND id not in(select id from fact_index)"""
+
+        rebuild_ids = ",".join([str(res[0]) for res in self.fetchall(index_query, (start_date, end_date))])
+
+        if rebuild_ids:
+            query = """
+                       SELECT a.id AS id,
+                              a.start_time AS start_time,
+                              a.end_time AS end_time,
+                              a.description as description,
+                              b.name AS name, b.id as activity_id,
+                              coalesce(c.name, ?) as category,
+                              e.name as tag
+                         FROM facts a
+                    LEFT JOIN activities b ON a.activity_id = b.id
+                    LEFT JOIN categories c ON b.category_id = c.id
+                    LEFT JOIN fact_tags d ON d.fact_id = a.id
+                    LEFT JOIN tags e ON e.id = d.tag_id
+                        WHERE a.id in (%s)
+                     ORDER BY a.id
+            """ % rebuild_ids
+
+            facts = self.__group_tags(self.fetchall(query, (_("Unsorted"), )))
+
+            insert = """INSERT INTO fact_index (id, name, category, description, tag)
+                             VALUES (?, ?, ?, ?, ?)"""
+            params = [(fact['id'], fact['name'], fact['category'], fact['description'], " ".join(fact['tags'])) for fact in facts]
+
+            self.executemany(insert, params)
+
 
     """ Here be dragons (lame connection/cursor wrappers) """
     def get_connection(self):
@@ -873,21 +885,32 @@ class Storage(storage.Storage):
         con = self.__con or self.connection
         cur = self.__cur or con.cursor()
 
-        if isinstance(statement, list) == False: #we kind of think that we will get list of instructions
+        if isinstance(statement, list) == False: # we expect to receive instructions in list
             statement = [statement]
             params = [params]
 
-        if isinstance(statement, list):
-            for i in range(len(statement)):
-                logging.debug("%s %s" % (statement[i], params[i]))
-
-                res = cur.execute(statement[i], params[i])
+        for state, param in zip(statement, params):
+            logging.debug("%s %s" % (state, param))
+            cur.execute(state, param)
 
         if not self.__con:
             con.commit()
             cur.close()
             self.register_modification()
 
+    def executemany(self, statement, params = []):
+        con = self.__con or self.connection
+        cur = self.__cur or con.cursor()
+
+        logging.debug("%s %s" % (statement, params))
+        cur.executemany(statement, params)
+
+        if not self.__con:
+            con.commit()
+            cur.close()
+            self.register_modification()
+
+
 
     def start_transaction(self):
         # will give some hints to execute not to close or commit anything
@@ -917,7 +940,7 @@ class Storage(storage.Storage):
 
         """upgrade DB to hamster version"""
         version = self.fetchone("SELECT version FROM version")["version"]
-        current_version = 7
+        current_version = 9
 
         if version < 2:
             """moving from fact_date, fact_time to start_time, end_time"""
@@ -1109,11 +1132,34 @@ class Storage(storage.Storage):
             self.execute("DROP table facts")
             self.execute("ALTER TABLE increment_facts RENAME TO facts")
 
+        if version < 8:
+            # working around sqlite's utf-f case sensitivity (bug 624438)
+            # more info: http://www.gsak.net/help/hs23820.htm
+            self.execute("ALTER TABLE activities ADD COLUMN search_name varchar2")
+
+            activities = self.fetchall("select * from activities")
+            statement = "update activities set search_name = ? where id = ?"
+            for activity in activities:
+                self.execute(statement, (activity['name'].lower(), activity['id']))
+
+            # same for categories
+            self.execute("ALTER TABLE categories ADD COLUMN search_name varchar2")
+            categories = self.fetchall("select * from categories")
+            statement = "update categories set search_name = ? where id = ?"
+            for category in categories:
+                self.execute(statement, (category['name'].lower(), category['id']))
+
+        if version < 9:
+            # adding full text search
+            self.execute("""CREATE VIRTUAL TABLE fact_index
+                                           USING fts3(id, name, category, description, tag)""")
+
 
         # at the happy end, update version number
         if version < current_version:
             #lock down current version
             self.execute("UPDATE version SET version = %d" % current_version)
+            print "updated database from version %d to %d" % (version, current_version)
 
         """we start with an empty database and then populate with default
            values. This way defaults can be localized!"""
